@@ -1,72 +1,47 @@
+from __future__ import annotations
+
 import ast
-import distutils.sysconfig as sysconfig
+import marshal
 import re
 import struct
 import sys
 from abc import ABC, abstractmethod
-from collections.abc import MutableSequence, Sequence
+from collections.abc import Iterable, Iterator, MutableSequence, Sequence
 from enum import Enum
 from io import BytesIO
-from pathlib import Path
 from pickletools import OpcodeInfo, genops, opcodes
 from typing import (
     Any,
     BinaryIO,
-    ByteString,
-    Dict,
-    FrozenSet,
     Generic,
-    Iterable,
-    Iterator,
-    List,
-    Optional,
-    Set,
-    Type,
     TypeVar,
-    Union,
     overload,
 )
-from typing import (
-    Tuple as TupleType,
-)
+
+from stdlib_list import in_stdlib
 
 from fickling.exception import WrongMethodError
 
 T = TypeVar("T")
 
-if sys.version_info < (3, 9):
-    # abstract collections were not subscriptable until Python 3.9
-    OpcodeSequence = MutableSequence
-    GenericSequence = Sequence
-
-    def make_constant(*args, **kwargs) -> ast.Constant:
-        # prior to Python 3.9, the ast.Constant class did not have a `kind` member, but the
-        # `astunparse` module expects that!
-        ret = ast.Constant(*args, **kwargs)
-        if not hasattr(ret, "kind"):
-            setattr(ret, "kind", None)
-        return ret
-
+if sys.version_info < (3, 12):
+    from typing_extensions import Buffer
 else:
-    OpcodeSequence = MutableSequence["Opcode"]
-    GenericSequence = Sequence[T]
-    make_constant = ast.Constant
+    from collections.abc import Buffer
 
-BUILTIN_MODULE_NAMES: FrozenSet[str] = frozenset(sys.builtin_module_names)
 
-OPCODES_BY_NAME: Dict[str, Type["Opcode"]] = {}
-OPCODE_INFO_BY_NAME: Dict[str, OpcodeInfo] = {opcode.name: opcode for opcode in opcodes}
+OpcodeSequence = MutableSequence["Opcode"]
+GenericSequence = Sequence[T]
+make_constant = ast.Constant
 
-STD_LIB = sysconfig.get_python_lib(standard_lib=True)
+BUILTIN_MODULE_NAMES: frozenset[str] = frozenset(sys.builtin_module_names)
+
+OPCODES_BY_NAME: dict[str, type[Opcode]] = {}
+OPCODE_INFO_BY_NAME: dict[str, OpcodeInfo] = {opcode.name: opcode for opcode in opcodes}
 
 
 def is_std_module(module_name: str) -> bool:
-    base_path = Path(STD_LIB).joinpath(*module_name.split("."))
-    return (
-        base_path.is_dir()
-        or base_path.with_suffix(".py").is_file()
-        or module_name in BUILTIN_MODULE_NAMES
-    )
+    return in_stdlib(module_name) or module_name in BUILTIN_MODULE_NAMES
 
 
 class MarkObject:
@@ -79,11 +54,11 @@ class Opcode:
 
     def __init__(
         self,
-        argument: Optional[Any] = None,
-        position: Optional[int] = None,
-        data: Optional[bytes] = None,
+        argument: Any | None = None,
+        position: int | None = None,
+        data: bytes | None = None,
         *,
-        info: Optional[OpcodeInfo] = None,
+        info: OpcodeInfo | None = None,
     ):
         if self.__class__ is Opcode:
             if info is None:
@@ -94,8 +69,8 @@ class Opcode:
                 f"{info!r}"
             )
         self.arg: Any = argument
-        self.pos: Optional[int] = position
-        self._data: Optional[bytes] = data
+        self.pos: int | None = position
+        self._data: bytes | None = data
 
     def has_data(self) -> bool:
         return self._data is not None
@@ -104,8 +79,7 @@ class Opcode:
     def data(self) -> bytes:
         if self._data is None:
             return self.encode()
-        else:
-            return self._data
+        return self._data
 
     @data.setter
     def data(self, value: bytes):
@@ -128,16 +102,14 @@ class Opcode:
         if cls is Opcode:
             if "info" not in kwargs:
                 raise ValueError(f"You must provide an `info` argument to construct {cls.__name__}")
-            else:
-                info = kwargs["info"]
+            info = kwargs["info"]
             if info.name in OPCODES_BY_NAME:
                 del kwargs["info"]
                 return OPCODES_BY_NAME[info.name](*args, **kwargs)
-            else:
-                raise NotImplementedError(f"TODO: Add support for Opcode {info.name}")
+            raise NotImplementedError(f"TODO: Add support for Opcode {info.name}")
         return super().__new__(cls)
 
-    def run(self, interpreter: "Interpreter"):
+    def run(self, interpreter: Interpreter):
         raise NotImplementedError(f"TODO: Add support for Pickle opcode {self.info.name}")
 
     def __init_subclass__(cls, **kwargs):
@@ -150,12 +122,12 @@ class Opcode:
         ):
             if not hasattr(cls, "name") or cls.name is None:
                 raise TypeError("Opcode subclasses must define a name")
-            elif cls.name in OPCODES_BY_NAME:
+            if cls.name in OPCODES_BY_NAME:
                 raise TypeError(f"An Opcode named {cls.name} is already defined")
-            elif cls.name not in OPCODE_INFO_BY_NAME:
+            if cls.name not in OPCODE_INFO_BY_NAME:
                 raise TypeError(f"An Opcode named {cls.name} is not defined in `pickletools`")
             OPCODES_BY_NAME[cls.name] = cls
-            setattr(cls, "info", OPCODE_INFO_BY_NAME[cls.name])
+            cls.info = OPCODE_INFO_BY_NAME[cls.name]
             # find the associated `pickletools` OpcodeInfo:
         return super().__init_subclass__(**kwargs)
 
@@ -228,7 +200,7 @@ class DynamicLength(Opcode, ABC):
 
 
 class NoOp(Opcode):
-    def run(self, interpreter: "Interpreter"):
+    def run(self, interpreter: Interpreter):
         pass
 
 
@@ -251,10 +223,10 @@ def raw_unicode_escape(byte_string: bytes) -> str:
 
 
 class ConstantOpcode(Opcode):
-    ConstantOpcodePriorities: Dict[Type["ConstantOpcode"], int] = {}
+    ConstantOpcodePriorities: dict[type[ConstantOpcode], int] = {}
     priority: int
 
-    def run(self, interpreter: "Interpreter"):
+    def run(self, interpreter: Interpreter):
         interpreter.stack.append(make_constant(self.arg))
 
     def __init_subclass__(cls, **kwargs):
@@ -262,7 +234,7 @@ class ConstantOpcode(Opcode):
         if not cls.__name__ == "ConstantInt":
             if cls.validate.__code__ == ConstantOpcode.validate.__code__:
                 raise TypeError(f"{cls.__name__} must implement the validate method")
-            elif (
+            if (
                 not hasattr(cls, "priority")
                 or not isinstance(cls.priority, int)
                 or cls.priority is None
@@ -285,7 +257,7 @@ class ConstantOpcode(Opcode):
         raise NotImplementedError()
 
     @classmethod
-    def new(cls: Type[T], obj) -> T:
+    def new(cls: type[T], obj) -> T:
         for subclass, _ in sorted(
             ConstantOpcode.ConstantOpcodePriorities.items(), key=lambda kv: kv[1]
         ):
@@ -324,18 +296,18 @@ class ConstantInt(ConstantOpcode, ABC):
         st = self.struct_types[self.num_bytes]
         if not self.signed:
             st = st.upper()
-        return struct.pack(f"{self.endianness.value}{st}")
+        return struct.pack(f"{self.endianness.value}{st}", self.arg)
 
     @classmethod
     def validate(cls, obj):
         if not isinstance(obj, int):
             raise ValueError(f"{cls.__name__} can only be instantiated from integers, not {obj!r}")
-        elif cls.num_bytes not in cls.struct_types:
+        if cls.num_bytes not in cls.struct_types:
             raise TypeError(
                 f"{cls.__name__}.struct_types does not include a value for "
                 f"{cls.__name__}.length_bytes = {cls.num_bytes}"
             )
-        elif obj < cls.min_value or obj > cls.max_value:
+        if obj < cls.min_value or obj > cls.max_value:
             raise ValueError(
                 f"Invalid value {obj!r}: {cls.__name__} can only represent lengths in the range "
                 f"[{cls.min_value}, {cls.max_value}]"
@@ -344,14 +316,14 @@ class ConstantInt(ConstantOpcode, ABC):
 
 
 class StackSliceOpcode(Opcode):
-    def run(self, interpreter: "Interpreter", stack_slice: List[ast.expr]):
+    def run(self, interpreter: Interpreter, stack_slice: list[ast.expr]):
         raise NotImplementedError(f"{self.__class__.__name__} must implement run()")
 
     def __init_subclass__(cls, **kwargs):
         ret = super().__init_subclass__(**kwargs)
         orig_run = cls.run
 
-        def run_wrapper(self, interpreter: "Interpreter"):
+        def run_wrapper(self, interpreter: Interpreter):
             args = []
             while True:
                 if not interpreter.stack:
@@ -359,35 +331,34 @@ class StackSliceOpcode(Opcode):
                 obj = interpreter.stack.pop()
                 if isinstance(obj, MarkObject):
                     break
-                else:
-                    args.append(obj)
+                args.append(obj)
             args = list(reversed(args))
             return orig_run(self, interpreter, args)
 
-        setattr(cls, "run", run_wrapper)
+        cls.run = run_wrapper
 
         return ret
 
 
 class ASTProperties(ast.NodeVisitor):
     def __init__(self):
-        self.imports: List[Union[ast.Import, ast.ImportFrom]] = []
-        self.calls: List[ast.Call] = []
-        self.non_setstate_calls: List[ast.Call] = []
-        self.likely_safe_imports: Set[str] = set()
+        self.imports: list[ast.Import | ast.ImportFrom] = []
+        self.calls: list[ast.Call] = []
+        self.non_setstate_calls: list[ast.Call] = []
+        self.likely_safe_imports: set[str] = set()
 
-    def _process_import(self, node: Union[ast.Import, ast.ImportFrom]):
+    def _process_import(self, node: ast.Import | ast.ImportFrom):
         self.imports.append(node)
         if isinstance(node, ast.ImportFrom) and is_std_module(node.module):
             self.likely_safe_imports |= {name.name for name in node.names}
 
-    def visit_Import(self, node: ast.Import):  # noqa: N802
+    def visit_Import(self, node: ast.Import):
         self._process_import(node)
 
-    def visit_ImportFrom(self, node: ast.ImportFrom):  # noqa: N802
+    def visit_ImportFrom(self, node: ast.ImportFrom):
         self._process_import(node)
 
-    def visit_Call(self, node: ast.Call):  # noqa: N802
+    def visit_Call(self, node: ast.Call):
         self.calls.append(node)
         if not isinstance(node.func, ast.Attribute) or node.func.attr != "__setstate__":
             self.non_setstate_calls.append(node)
@@ -402,10 +373,13 @@ class EmptyPickleError(PickleDecodeError):
 
 
 class Pickled(OpcodeSequence):
-    def __init__(self, opcodes: Iterable[Opcode]):
-        self._opcodes: List[Opcode] = list(opcodes)
-        self._ast: Optional[ast.Module] = None
-        self._properties: Optional[ASTProperties] = None
+    def __init__(self, opcodes: Iterable[Opcode], has_invalid_opcode: bool = False):
+        self._opcodes: list[Opcode] = list(opcodes)
+        self._ast: ast.Module | None = None
+        self._properties: ASTProperties | None = None
+        # Whether the pickled sequence was interrupted because of
+        # an invalid opcode
+        self._has_invalid_opcode: bool = has_invalid_opcode
 
     def __len__(self) -> int:
         return len(self._opcodes)
@@ -421,6 +395,46 @@ class Pickled(OpcodeSequence):
         self._ast = None
         self._properties = None
 
+    def _is_constant_type(self, obj: Any) -> bool:
+        return isinstance(obj, (int, float, str, bytes))
+
+    def _encode_python_obj(self, obj: Any) -> List[Opcode]:
+        """Create an opcode sequence that builds an arbitrary python object on the top of the
+        pickle VM stack"""
+        if self._is_constant_type(obj):
+            return [ConstantOpcode.new(obj)]
+        if isinstance(obj, list):
+            res = [Mark()]
+            for item in obj:
+                if self._is_constant_type(item):
+                    res.append(ConstantOpcode.new(item))
+                else:
+                    res += self._encode_python_obj(item)
+            res.append(List())
+            return res
+        if isinstance(obj, dict):
+            if len(obj) == 0:
+                res = [EmptyDict()]
+            else:
+                res = [Mark()]
+                for key, val in obj.items():
+                    res.append(ConstantOpcode.new(key))  # Assume key is constant
+                    if self._is_constant_type(val):
+                        res.append(ConstantOpcode.new(val))
+                    else:
+                        res += self._encode_python_obj(val)
+                res.append(Dict())
+            return res
+        raise ValueError(f"Type {type(obj)} not supported")
+
+    def insert_python_obj(self, index: int, obj: Any) -> int:
+        """Insert an opcode sequence that constructs a python object on the stack.
+        Returns the number of opcodes inserted"""
+        opcodes = self._encode_python_obj(obj)
+        for i, opcode in enumerate(opcodes):
+            self.insert(index + i, opcode)
+        return len(opcodes)
+
     def insert_python(
         self,
         *args,
@@ -446,38 +460,48 @@ class Pickled(OpcodeSequence):
         self.insert(i, Mark())
         i += 1
         for arg in args:
-            self.insert(i, ConstantOpcode.new(arg))
-            i += 1
+            i += self.insert_python_obj(i, arg)
+            # self.insert(i, ConstantOpcode.new(arg))
+            # i += 1
         self.insert(i, Tuple())
         i += 1
         if run_first:
             self.insert(i, Reduce())
             if use_output_as_unpickle_result:
-                self.insert(-1, Pop())
-            return i + 1
-        else:
-            if use_output_as_unpickle_result:
-                # the top of the stack should be the original unpickled value, but we can throw
-                # that away because we are replacing it with the result of calling eval:
-                self.insert(-1, Pop())
-                # now the top of the stack should be our original Global, Mark, Unicode,
-                # Tuple setup, ready for Reduce:
-                self.insert(-1, Reduce())
+                self.insert(-1, Pop())  # Just discard the original unpickle
             else:
-                # we need to preserve the "real" output of the preexisting unpickling, which should
-                # be at the top of the stack, directly above our Tuple, Unicode, Mark, and Global
-                # stack items we added above.
-                # So, we have to save the original result to the memo. First, interpret the existing
-                # code to see which memo location it would be saved to:
-                interpreter = Interpreter(self)
-                interpreter.run()
-                memo_id = len(interpreter.memory)
-                self.insert(-1, Memoize())
+                # At the end, the stack will contain [reduce_res, original_obj].
+                # We need to remove everything below original_obj:
+                # NOTE(boyan): using an arbitrary MEMO key here seems to work. If not,
+                # then switch to using the interpreter() to determine the correct MEMO key to use here
+                self.insert(-1, Put(321987))  # Put obj in memo
+                self.insert(-1, Pop())  # Pop obj and reduce_res under
                 self.insert(-1, Pop())
-                self.insert(-1, Reduce())
-                self.insert(-1, Pop())
-                self.insert(-1, Get.create(memo_id))
-            return -1
+                self.insert(-1, Get.create(321987))  # Get back obj
+            return i + 1
+        # Inject call
+        if use_output_as_unpickle_result:
+            # the top of the stack should be the original unpickled value, but we can throw
+            # that away because we are replacing it with the result of calling eval:
+            self.insert(-1, Pop())
+            # now the top of the stack should be our original Global, Mark, Unicode,
+            # Tuple setup, ready for Reduce:
+            self.insert(-1, Reduce())
+        else:
+            # we need to preserve the "real" output of the preexisting unpickling, which should
+            # be at the top of the stack, directly above our Tuple, Unicode, Mark, and Global
+            # stack items we added above.
+            # So, we have to save the original result to the memo. First, interpret the existing
+            # code to see which memo location it would be saved to:
+            interpreter = Interpreter(self)
+            interpreter.run()
+            memo_id = len(interpreter.memory)
+            self.insert(-1, Memoize())
+            self.insert(-1, Pop())
+            self.insert(-1, Reduce())
+            self.insert(-1, Pop())
+            self.insert(-1, Get.create(memo_id))
+        return -1
 
     insert_python_eval = insert_python
 
@@ -519,7 +543,8 @@ class Pickled(OpcodeSequence):
     def insert_function_call_on_unpickled_object(
         self,
         function_definition: str,
-        constant_args: Optional[List[Any]] = None,
+        constant_args: List[Any] | None = None,
+        compile_code: bool = False,
     ):
         """Insert and call a function that takes the unpickled object as parameter.
 
@@ -527,6 +552,11 @@ class Pickled(OpcodeSequence):
         to call, including the `def` keyword. The function prototype must be `myfunc(obj)` where
         `obj` is the object being unpickled. The function return value is used as the unpickling
         output.
+
+        :param compile_code: whether the function definition should be precompiled into
+        Python bytecode, for increased obfuscation. Note that the function name will still be
+        exposed as plaintext sourcecode as this is required to make the function callable through
+        a call to (pseudocode) "eval(function_name)".
         """
 
         if not isinstance(self[-1], Stop):
@@ -539,35 +569,76 @@ class Pickled(OpcodeSequence):
         function_name = fn_match[0]
 
         # Insert exec of the function definition in advance
-        self.append_python(function_definition, attr="exec", pop_result=True)
+        if compile_code:
+            # Compile the function definition
+            code_obj = compile(function_definition, "<string>", "exec")
+            bytecode = marshal.dumps(code_obj)  # marshal is required to get literal bytes
+
+            # Instructions:
+            ## Current stack status: [model]
+            ## Add the compiled bytes to the stack, then unmarshal them back into a code object
+            self.append_python(bytecode, module="marshal", attr="loads", pop_result=False)
+            ## [model, def func]
+            ## Slide exec instructions under the function definition on the stack
+            self.insert(-1, Put(1))  # Move function def off the stack and into memory
+            self.insert(-1, Pop())
+            ## [model]
+            if not isinstance(self[-1], Stop):  # sanity check akin to the one in append_python
+                raise ValueError("Expected the last opcode to be STOP")
+            ### NOTE(boyan): this seems to work even without insert GLOBAL at the beginning
+            ### of the pickle, but see comment in 'insert_python'
+            self.insert(-1, Global.create("builtins", "exec"))
+            ## [model, exec]
+            self.insert(-1, Mark())
+            ## [model, exec, mark]
+            self.insert(-1, Get.create(1))  # Place the function def back
+            ## [model, exec, mark, def func]
+            self.insert(-1, Tuple())
+            ## [model, exec, (def func)]
+            self.insert(-1, Reduce())
+            ## [model, exec return value]
+            self.insert(-1, Pop())  # Remove extraneous return value from the stack
+            ## [model]
+        else:
+            ## Current stack status: [model]
+            self.append_python(function_definition, attr="exec", pop_result=True)
+            ## [model]
 
         # Eval the function name get the callable object
         # If we inject myfunc() this will return eval(myfunc) which is the myfunc callable object
         self.append_python(function_name, attr="eval")
+        # [model, func]
 
-        # At the end of exec, the stack contains [model, func]. We
-        # swap them on the stack
+        # At the end of all of the above operations, the stack contains [model, func].
+        # We swap them on the stack:
         self.insert(-1, Put(1))  # Put func in memo
         self.insert(-1, Pop())
+        # [model]
         self.insert(-1, Put(2))  # Put model in memo
         self.insert(-1, Pop())
+        # []
         self.insert(-1, Get.create(1))
+        # [func]
         self.insert(-1, Mark())
+        # [func, mark]
         self.insert(-1, Get.create(2))
+        # [func, mark, model]
 
         # Add constant arguments
         if constant_args:
             for arg in constant_args:
                 self.insert(-1, ConstantOpcode.new(arg))
+        # [func, mark, model, arg1, ..., argn]
 
-        # Now the stack contains [func, mark, model].
         # We need to add TUPLE which
         # packs the function arguments from the stack and then call REDUCE, which calls the injected
         # function.
         # Note: precondition says the function must return the final object so no need to save the
         # object before calling reduce.
         self.insert(-1, Tuple())
+        # [func, (model, arg1, ..., argn)]
         self.insert(-1, Reduce())
+        # [func return value]
 
     def insert_python_exec(
         self,
@@ -583,7 +654,7 @@ class Pickled(OpcodeSequence):
             use_output_as_unpickle_result=use_output_as_unpickle_result,
         )
 
-    def __setitem__(self, index: Union[int, slice], item: Union[Opcode, Iterable[Opcode]]):
+    def __setitem__(self, index: int | slice, item: Opcode | Iterable[Opcode]):
         self._opcodes[index] = item
         self._ast = None
         self._properties = None
@@ -621,19 +692,31 @@ class Pickled(OpcodeSequence):
     def opcodes(self) -> Iterator[Opcode]:
         return iter(self)
 
+    @property
+    def has_invalid_opcode(self) -> bool:
+        return self._has_invalid_opcode
+
     @staticmethod
-    def make_stream(data: Union[ByteString, BinaryIO]) -> BinaryIO:
-        if isinstance(data, (bytes, bytearray, ByteString)):
+    def make_stream(data: Buffer | BinaryIO) -> BinaryIO:
+        if isinstance(data, (bytes, bytearray, Buffer)):
             data = BytesIO(data)
         elif (not hasattr(data, "seekable") or not data.seekable()) and hasattr(data, "read"):
             data = BytesIO(data.read())
         return data
 
     @staticmethod
-    def load(pickled: Union[ByteString, BinaryIO]) -> "Pickled":
+    def load(pickled: Buffer | BinaryIO, fail_on_decode_error: bool = True) -> Pickled:
+        """
+        :arg fail_on_decode_error: If a decoding error such as unknown opcode happens
+            and this parameter is set to False, then the error is ignored and the
+            returns the list of opcodes decoded until the error. If set to True,
+            then raises a decode error exception
+        """
+
         pickled = Pickled.make_stream(pickled)
         first_pos = pickled.tell()
         opcodes: List[Opcode] = []
+        has_invalid_opcode = False
 
         try:
             for info, arg, pos in genops(pickled):
@@ -671,7 +754,9 @@ class Pickled(OpcodeSequence):
                     pickled.seek(pos_before)
         except ValueError as e:
             if opcodes:
-                raise PickleDecodeError(e)
+                if fail_on_decode_error:
+                    raise PickleDecodeError(e)
+                has_invalid_opcode = True
             else:
                 raise EmptyPickleError()
         if opcodes:
@@ -683,7 +768,7 @@ class Pickled(OpcodeSequence):
                 pickled.seek(last_pos)
         else:
             pickled.seek(first_pos)
-        return Pickled(opcodes)
+        return Pickled(opcodes, has_invalid_opcode=has_invalid_opcode)
 
     @property
     def properties(self) -> ASTProperties:
@@ -719,7 +804,7 @@ on the Pickled object instead"""
             "This method has been removed. Use fickling.is_likely_safe() on the pickle file instead"
         )
 
-    def unsafe_imports(self) -> Iterator[Union[ast.Import, ast.ImportFrom]]:
+    def unsafe_imports(self) -> Iterator[ast.Import | ast.ImportFrom]:
         for node in self.properties.imports:
             if node.module in (
                 "__builtin__",
@@ -737,7 +822,7 @@ on the Pickled object instead"""
             elif "eval" in (n.name for n in node.names):
                 yield node
 
-    def non_standard_imports(self) -> Iterator[Union[ast.Import, ast.ImportFrom]]:
+    def non_standard_imports(self) -> Iterator[ast.Import | ast.ImportFrom]:
         for node in self.properties.imports:
             if not is_std_module(node.module):
                 yield node
@@ -756,7 +841,7 @@ on the Pickled object instead"""
 class Stack(GenericSequence, Generic[T]):
     def __init__(self, initial_value: Iterable[T] = ()):
         self._stack: List[T] = list(initial_value)
-        self.opcode: Optional[Opcode] = None
+        self.opcode: Opcode | None = None
 
     @overload
     @abstractmethod
@@ -776,10 +861,8 @@ class Stack(GenericSequence, Generic[T]):
         if not self._stack:
             if self.opcode is None:
                 raise IndexError("Stack is empty")
-            else:
-                raise IndexError(f"Opcode {self.opcode!s} attempted to pop from an empty stack")
-        else:
-            return self._stack.pop()
+            raise IndexError(f"Opcode {self.opcode!s} attempted to pop from an empty stack")
+        return self._stack.pop()
 
     def push(self, obj: T):
         self._stack.append(obj)
@@ -794,7 +877,7 @@ class Stack(GenericSequence, Generic[T]):
 
 
 class ModuleBody:
-    def __init__(self, interpreter: "Interpreter"):
+    def __init__(self, interpreter: Interpreter):
         self._list: List[ast.stmt] = []
         self.interpreter: Interpreter = interpreter
 
@@ -805,7 +888,7 @@ class ModuleBody:
                 f"Statement {stmt} was expected to have line number {lineno} but instead has "
                 f"{stmt.lineno}"
             )
-        setattr(stmt, "lineno", lineno)
+        stmt.lineno = lineno
         self._list.append(stmt)
 
     def extend(self, stmts: Iterable[ast.stmt]):
@@ -818,7 +901,7 @@ class ModuleBody:
     def __len__(self):
         return len(self._list)
 
-    def __getitem__(self, index: Union[int, slice]) -> ast.stmt:
+    def __getitem__(self, index: int | slice) -> ast.stmt:
         return self._list[index]
 
 
@@ -827,11 +910,11 @@ class Interpreter:
         self, pickled: Pickled, first_variable_id: int = 0, result_variable: str = "result"
     ):
         self.pickled: Pickled = pickled
-        self.memory: Dict[int, ast.expr] = {}
-        self.stack: Stack[Union[ast.expr, MarkObject]] = Stack()
+        self.memory: dict[int, ast.expr] = {}
+        self.stack: Stack[ast.expr | MarkObject] = Stack()
         self.module_body: ModuleBody = ModuleBody(self)
         self.result_variable: str = result_variable
-        self._module: Optional[ast.Module] = None
+        self._module: ast.Module | None = None
         self._var_counter: int = first_variable_id
         self._opcodes: Iterator[Opcode] = iter(pickled)
 
@@ -844,12 +927,12 @@ class Interpreter:
             self.run()
         return self._module
 
-    def unused_assignments(self) -> Dict[str, ast.Assign]:
+    def unused_assignments(self) -> dict[str, ast.Assign]:
         if self._module is None:
             self.run()
-        used: Set[str] = set()
-        defined: Set[str] = set()
-        assignments: Dict[str, ast.Assign] = {}
+        used: set[str] = set()
+        defined: set[str] = set()
+        assignments: dict[str, ast.Assign] = {}
         for statement in self.module_body:
             # skip the last statement because it is always used
             if isinstance(statement, ast.Assign):
@@ -876,7 +959,7 @@ class Interpreter:
                         used.add(node.id)
         return {varname: assignments[varname] for varname in defined - used}
 
-    def unused_variables(self) -> FrozenSet[str]:
+    def unused_variables(self) -> frozenset[str]:
         return self.unused_assignments().keys()  # type: ignore
 
     def stop(self):
@@ -898,15 +981,15 @@ class Interpreter:
             finished = True
         if finished:
             for i, stmt in enumerate(self.module_body):
-                setattr(stmt, "lineno", i + 1)
-                setattr(stmt, "col_offset", 0)
+                stmt.lineno = i + 1
+                stmt.col_offset = 0
             self._module = ast.Module(list(self.module_body), type_ignores=[])
             raise StopIteration()
         self.stack.opcode = opcode
         opcode.run(self)
         return opcode
 
-    def new_variable(self, value: ast.expr, name: Optional[str] = None) -> str:
+    def new_variable(self, value: ast.expr, name: str | None = None) -> str:
         if name is None:
             name = f"_var{self._var_counter}"
             self._var_counter += 1
@@ -925,7 +1008,7 @@ class Proto(NoOp):
     name = "PROTO"
 
     @staticmethod
-    def create(version: int) -> "Proto":
+    def create(version: int) -> Proto:
         return Proto(version)
 
     def encode_body(self) -> bytes:
@@ -935,18 +1018,17 @@ class Proto(NoOp):
     def version(self) -> int:
         if self.arg is None:
             return 0
-        elif isinstance(self.arg, int):
+        if isinstance(self.arg, int):
             return self.arg
-        else:
-            # Endianness shouldn't really matter here because there is only one byte for the version
-            return int.from_bytes(self.arg, "big", signed=False)
+        # Endianness shouldn't really matter here because there is only one byte for the version
+        return int.from_bytes(self.arg, "big", signed=False)
 
 
 class Global(Opcode):
     name = "GLOBAL"
 
     @staticmethod
-    def create(module: str, attr: str) -> "Global":
+    def create(module: str, attr: str) -> Global:
         return Global(f"{module} {attr}")
 
     @property
@@ -964,11 +1046,7 @@ class Global(Opcode):
             # no need to emit an import for builtins!
             pass
         else:
-            if sys.version_info < (3, 9):
-                # workaround for a bug in astunparse
-                alias = ast.alias(attr, asname=None)
-            else:
-                alias = ast.alias(attr)
+            alias = ast.alias(attr)
             interpreter.module_body.append(ast.ImportFrom(module=module, names=[alias], level=0))
         interpreter.stack.append(ast.Name(attr, ast.Load()))
 
@@ -986,19 +1064,58 @@ class StackGlobal(NoOp):
             module = module.value
         if isinstance(attr, ast.Constant):
             attr = attr.value
+
+        # normalize module and attr to strings
+        if not isinstance(module, str) or not isinstance(attr, str):
+            sys.stdout.write(
+                f"Warning: malformed pickle file. STACK_GLOBAL called with invalid types. "
+                f"'Module' is {type(module).__name__} ({module!r}), 'Attr' is {type(attr).__name__} ({attr!r}). "
+                f"Expected str; casting to string to continue analysis.\n"
+            )
+            module = str(module)
+            attr = str(attr)
         if module in ("__builtin__", "__builtins__", "builtins"):
             # no need to emit an import for builtins!
             pass
         else:
-            if sys.version_info < (3, 9):
-                # workaround for a bug in astunparse
-                alias = ast.alias(attr, asname=None)
-            else:
-                alias = ast.alias(attr)
+            alias = ast.alias(attr)
             interpreter.module_body.append(ast.ImportFrom(module=module, names=[alias], level=0))
         interpreter.stack.append(ast.Name(attr, ast.Load()))
 
 
+class Inst(StackSliceOpcode):
+    name = "INST"
+
+    @staticmethod
+    def create(module: str, classname: str) -> Inst:
+        return Inst(f"{module} {classname}")
+
+    @property
+    def module(self) -> str:
+        return next(iter(self.arg.split(" ")))
+
+    @property
+    def cls(self) -> str:
+        _, classname, *_ = self.arg.split(" ")
+        return classname
+
+    def run(self, interpreter: Interpreter, stack_slice: List[ast.expr]):
+        module, classname = self.module, self.cls
+        if module in ("__builtin__", "__builtins__", "builtins"):
+            # no need to emit an import for builtins!
+            pass
+        else:
+            alias = ast.alias(classname)
+            interpreter.module_body.append(ast.ImportFrom(module=module, names=[alias], level=0))
+        args = ast.Tuple(tuple(stack_slice))
+        call = ast.Call(ast.Name(classname, ast.Load()), list(args.elts), [])
+        var_name = interpreter.new_variable(call)
+        interpreter.stack.append(ast.Name(var_name, ast.Load()))
+
+    def encode(self) -> bytes:
+        return f"i{self.module}\n{self.classname}\n".encode()
+
+
 class Put(Opcode):
     name = "PUT"
 
@@ -1241,6 +1358,10 @@ class BinPersId(Opcode):
         )
 
 
+class PersId(Opcode):
+    name = "PERSID"
+
+
 class NoneOpcode(Opcode):
     name = "NONE"
 
@@ -1292,14 +1413,28 @@ class BinGet(Opcode):
     name = "BINGET"
 
     def run(self, interpreter: Interpreter):
-        interpreter.stack.append(interpreter.memory[self.arg])
+        if self.arg not in interpreter.memory:
+            sys.stderr.write(
+                f"Warning: malformed pickle file. BINGET references non-existent memo key {self.arg}; "
+                f"using placeholder None value to continue analysis\n"
+            )
+            interpreter.stack.append(ast.Constant(value=None))
+        else:
+            interpreter.stack.append(interpreter.memory[self.arg])
 
 
 class LongBinGet(Opcode):
     name = "LONG_BINGET"
 
     def run(self, interpreter: Interpreter):
-        interpreter.stack.append(interpreter.memory[self.arg])
+        if self.arg not in interpreter.memory:
+            sys.stderr.write(
+                f"Warning: malformed pickle file. LONG_BINGET references non-existent memo key {self.arg}; "
+                f"using placeholder None value to continue analysis\n"
+            )
+            interpreter.stack.append(ast.Constant(value=None))
+        else:
+            interpreter.stack.append(interpreter.memory[self.arg])
 
 
 class Get(Opcode):
@@ -1310,13 +1445,20 @@ class Get(Opcode):
         return int(self.arg)
 
     def run(self, interpreter: Interpreter):
-        interpreter.stack.append(interpreter.memory[self.memo_id])
+        if self.memo_id not in interpreter.memory:
+            sys.stderr.write(
+                f"Warning: malformed pickle file. BINGET references non-existent memo key {self.memo_id}; "
+                f"using placeholder None value to continue analysis\n"
+            )
+            interpreter.stack.append(ast.Constant(value=None))
+        else:
+            interpreter.stack.append(interpreter.memory[self.memo_id])
 
     def encode_body(self) -> bytes:
         return f"{self.memo_id}\n".encode()
 
     @staticmethod
-    def create(memo_id: int) -> "Get":
+    def create(memo_id: int) -> Get:
         return Get(f"{memo_id}\n".encode())
 
 
@@ -1581,16 +1723,12 @@ class Dict(Opcode):
         interpreter.stack.append(ast.Dict(keys=reversed(keys), values=reversed(values)))
 
 
-if sys.version_info < (3, 9):
-    # abstract collections were not subscriptable until Python 3.9
-    PickledSequence = Sequence
-else:
-    PickledSequence = Sequence[Pickled]
+PickledSequence = Sequence[Pickled]
 
 
 class StackedPickle(PickledSequence):
     def __init__(self, pickled: Iterable[Pickled]):
-        self.pickled: TupleType[Pickled, ...] = tuple(pickled)
+        self.pickled: tuple[Pickled, ...] = tuple(pickled)
 
     def __getitem__(self, index: int) -> Pickled:
         return self.pickled[index]
@@ -1599,12 +1737,12 @@ class StackedPickle(PickledSequence):
         return len(self.pickled)
 
     @staticmethod
-    def load(pickled: Union[ByteString, BinaryIO]) -> "StackedPickle":
+    def load(pickled: Buffer | BinaryIO, fail_on_decode_error: bool = True) -> StackedPickle:
         pickled = Pickled.make_stream(pickled)
         pickles: List[Pickled] = []
         while True:
             try:
-                p = Pickled.load(pickled)
+                p = Pickled.load(pickled, fail_on_decode_error=fail_on_decode_error)
                 if len(p) == 0:
                     break
                 pickles.append(p)
