@@ -1,9 +1,13 @@
+import ast
 import os
 import shutil
-import sys
+import struct
 import tarfile
+import tempfile
 import zipfile
 
+import numpy.lib.format as npformat
+
 from fickling.fickle import Pickled, StackedPickle
 
 """
@@ -46,8 +50,7 @@ def check_and_find_in_zip(
                     return any(
                         entry.endswith(file_name_or_extension) for entry in zip_file.namelist()
                     )
-                else:
-                    return any(file_name_or_extension in entry for entry in zip_file.namelist())
+                return any(file_name_or_extension in entry for entry in zip_file.namelist())
         else:
             return next(
                 (entry for entry in zip_path.namelist() if entry.endswith(file_name_or_extension)),
@@ -61,11 +64,44 @@ def check_and_find_in_zip(
         return None if return_path else False
 
 
-def check_pickle(file):
+def check_numpy(file):  # returns isNumpy,isNumpyPickle
+    """Checks if the numpy magic bytes are there, and if they are, if the header
+    claims the data is an object"""
+    file.seek(0)
+    try:
+        version = npformat.read_magic(file)
+    except ValueError:
+        return False, False  # not numpy
+
+    # This is a private variable, but the alternative
+    # would require using private functions or
+    # maintaining Numpy's version list in Fickling
+    hinfo = npformat._header_size_info.get(version)
+    if hinfo is None:
+        return False, False  # not a valid version of numpy
+    hlength_type, encoding = hinfo
+
+    hlength_str = file.read(struct.calcsize(hlength_type))
+    header_length = struct.unpack(hlength_type, hlength_str)[0]
+    header = file.read(header_length)
+    header = header.decode(encoding)
+
+    # The literal_eval can be abused to cause a DoS
+    # However this is also what Numpy uses,
+    # so it applies to np.load(fname, allow_pickle=False)
+    d = ast.literal_eval(header)
+    dtype = npformat.descr_to_dtype(d["descr"])
+
+    if dtype.hasobject:
+        return True, True  # numpy pickle
+    return True, False  # numpy non-pickle
+
+
+def check_pickle(file, min_length=0):
     """Checks if a file can be pickled; this does not directly determine the file is a pickle"""
     try:
-        Pickled.load(file)
-        return True
+        opcodes = Pickled.load(file).opcodes()
+        return len(opcodes) > min_length
     except Exception:  # noqa
         try:
             StackedPickle.load(file)
@@ -83,16 +119,21 @@ def find_file_properties(file_path, print_properties=False):
         properties["is_torch_zip"] = is_torch_zip
 
         # This tarfile check has many false positives. It is not a determinant of PyTorch v0.1.1.
-        if sys.version_info >= (3, 9):
-            is_tar = tarfile.is_tarfile(file)
-        else:
-            is_tar = tarfile.is_tarfile(file_path)
+        is_tar = tarfile.is_tarfile(file)
         properties["is_tar"] = is_tar
 
         # Similar to tar, this is not a robust verification.
-        is_valid_pickle = check_pickle(file)
+        # tar files often start with . which is a 1 operator pickle
+        is_valid_pickle = check_pickle(file, min_length=2)
         properties["is_valid_pickle"] = is_valid_pickle
 
+        # Numpy has a special header and magic bytes,
+        # mimics Numpy code to check if the file is Numpy
+        # and if it claims to contain a pickle
+        is_numpy, is_numpy_pickle = check_numpy(file)
+        properties["is_numpy"] = is_numpy
+        properties["is_numpy_pickle"] = is_numpy_pickle
+
         # PyTorch MAR can be a standard ZIP, but not a PyTorch ZIP
         # Some other non-PyTorch file formats rely on ZIP without PyTorch's limitations
         is_standard_zip = zipfile.is_zipfile(file)
@@ -128,6 +169,50 @@ def find_file_properties(file_path, print_properties=False):
     return properties
 
 
+def find_file_properties_recursively(file_path, print_properties=False):
+    """Property discovery that looks inside zip and tar archives"""
+    properties = find_file_properties(file_path, print_properties)
+
+    # check zip
+    check_zip = properties["is_standard_zip"]
+    # if it's the correct type of zip to be torch, make sure it's not a torch file
+    if check_zip and not properties["is_standard_not_torch"]:
+        for key in properties.keys():
+            if key.startswith("has_") and properties[key]:
+                # if it is a torch file, no need to check it
+                check_zip = False
+                break
+    # actually check the zip
+    if check_zip:
+        properties["children"] = {}
+        with tempfile.TemporaryDirectory() as tempdir:
+            with zipfile.ZipFile(file_path) as zipped_file:
+                for fname in zipped_file.namelist():
+                    zipped_file.extract(fname, path=tempdir)
+                    fname_path = os.path.join(tempdir, fname)
+                    properties["children"][fname] = find_file_properties_recursively(
+                        fname_path, print_properties
+                    )
+
+    # check tar
+    if properties["is_tar"]:  # tar archive
+        properties["children"] = {}
+        with tempfile.TemporaryDirectory() as tempdir:
+            with tarfile.TarFile(file_path) as tarred_file:
+                for fname in tarred_file.getnames():
+                    content = tarred_file.extractfile(fname)
+                    if content is None:
+                        properties["children"][fname] = None
+                        continue
+                    fname_path = os.path.join(tempdir, os.path.basename(fname))
+                    open(fname_path, "wb").write(content.read())
+                    properties["children"][fname] = find_file_properties_recursively(
+                        fname_path, print_properties
+                    )
+
+    return properties
+
+
 def check_if_legacy_format(file):
     """PyTorch v0.1.1: Tar file with sys_info, pickle, storages, and tensors"""
     required_entries = {"pickle", "storages", "tensors"}
